#ifndef MUSCLES_DESCENT_H
#define MUSCLES_DESCENT_H

#include <algorithm>
#include <vector>

struct Descent {
	template<typename Solution>
	Solution operator()(Solution const&initial) const {
		Solution solution{initial};
		auto cost = solution.value();

		std::vector<Solution> rejectedSolutions;

		bool stop = false;
		while(!stop) {
			Solution bestNeighbour;
			bool found = findBestNeighbour(solution, bestNeighbour, rejectedSolutions);

			if(found) {
				auto bestCost = bestNeighbour.value();
				if(bestCost <= cost) {
					rejectedSolutions.push_back(solution);

					solution = std::move(bestNeighbour);
					cost = std::move(bestCost);
				} else
					stop = true;
			} else
				stop = true;
		}

		return solution;
	}

private:
	template<typename Solution>
	bool findBestNeighbour(Solution const&solution, Solution &best, std::vector<Solution> const&rejectedSolutions) const {
		bool found = false;

		best = solution;

		for(std::size_t i = 0; i < solution.size()-1; ++i) {
			for(std::size_t j = i+1; j < solution.size(); ++j) {
				Solution neighbour{solution};
				neighbour.swap(i, j);

				typename std::decay_t<decltype(rejectedSolutions)>::const_iterator it;
				it = std::find(std::begin(rejectedSolutions), std::end(rejectedSolutions), neighbour);
				if(it == std::end(rejectedSolutions)) {
					if(neighbour.value() < best.value()) {
						best = neighbour;
						found = true;
					}
				}
			}
		}

		return found;
	}
};

#endif
