#ifndef MUSCLES_RGREEDY_H
#define MUSCLES_RGREEDY_H

#include <random>
#include <set>
#include <vector>

#include "../tsp/types.h"

template<typename Solution>
struct RGreedy {
	std::size_t n;

	RGreedy(std::size_t n = {}): n{n} {}

	template<typename Problem, typename RandomEngine>
	Solution operator()(Problem const&problem, RandomEngine &gen, std::size_t start = 0) const {
		Solution solution{problem};

		std::vector<std::size_t> l;
		l.reserve(problem.points.size()-1);
		for(std::size_t i = 0; i < problem.points.size(); ++i)
			if(i != start)
				l.push_back(i);
		solution.push_back(start);

		// types for the set
		using SetElement = std::pair<decltype(l)::iterator, tsp::Distance>;
		struct Comparator {
			bool operator()(SetElement const&a, SetElement const&b) const {
				return a.second < b.second;
			};
		};

		while(l.size()) {
			std::set<SetElement, Comparator> bests;

			// for each remaining point, sort by cost
			for(auto it = std::begin(l); it != std::end(l); ++it) {
				tsp::Distance cost = solution.valueWith(*it);
				bests.emplace(it, cost);
			}

			// add element to solution
			{
				std::size_t m = n<bests.size()? n:bests.size();
				decltype(SetElement::first) best = std::begin(bests)->first;

				// select one of n best elements
				if(m) {
					std::uniform_int_distribution<std::size_t> dist{0, m-1};

					std::size_t i = dist(gen);
					typename decltype(bests)::const_iterator it = std::begin(bests);
					std::advance(it, i);

					best = it->first;
				}

				solution.push_back(*best);
				l.erase(best);
			}
		}

		return solution;
	}
};

#endif
