#ifndef ROSA_ELS_H
#define ROSA_ELS_H

#include <alsk/alsk.h>

namespace rosa {

/* ELS
 * best = ls(s)             -- SEls
 * loop                     -- SElsOuterLoop
 * * loop                   -- SElsInnerLoop
 * * * s     = mutate(best) -- SElsGen
 * * * s     = ls(s)
 * * * ibest = select(s, ibest)
 * * ----
 * * best = select(s, best) // with acceptation criteria?
 * ----
 * return best
 */
template<
	typename InitLS, typename Mutate, typename LS,
	typename InnerSelect, typename OuterSelect
>
using SkelElsStruct =
alsk::S<alsk::Serial,
	InitLS,
	alsk::S<alsk::IterSel,
		alsk::S<alsk::FarmSel,
			alsk::S<alsk::Serial,
				Mutate, LS
			>,
			InnerSelect
		>,
		OuterSelect
	>
>;

template<typename Solution>
using SkelElsLinks =
alsk::L<alsk::Serial, alsk::arg::R<1>(Solution const&),
	Solution(alsk::arg::P<0>),
	alsk::L<alsk::IterSel, Solution(alsk::arg::R<0> const&),
		alsk::L<alsk::FarmSel, Solution(Solution),
			alsk::L<alsk::Serial, alsk::arg::R<1>(alsk::arg::P<0>),
				Solution(alsk::arg::P<0>, alsk::arg::RNG&),
				Solution(alsk::arg::R<0> const&)
			>,
			Solution(Solution const&, Solution const&)
		>,
		Solution(Solution const&, Solution const&)
	>
>;

template<
	typename Solution,
	typename InitLS, typename Mutate, typename LS,
	typename InnerSelect, typename OuterSelect = InnerSelect
>
using SkelEls = alsk::BuildSkeleton<SkelElsStruct, SkelElsLinks>::skeleton<
	tmp::Pack<InitLS, Mutate, LS, InnerSelect, OuterSelect>,
	tmp::Pack<Solution>
>;

}

#endif
