#ifndef ROSA_GRASP_H
#define ROSA_GRASP_H

#include <alsk/alsk.h>

namespace rosa {

/* GRASP
 * loop
 * * s    = init()
 * * s    = ls(s)
 * * best = select(s, best)
 * ----
 * return best
 */
template<typename Init, typename LS, typename Select>
using SkelGraspStructure =
alsk::S<alsk::FarmSel,
	alsk::S<alsk::Serial, Init, LS>,
	Select
>;

template<typename Problem, typename Solution>
using SkelGraspLinks =
alsk::L<alsk::FarmSel, Solution(Problem const&),
	alsk::L<alsk::Serial, alsk::arg::R<1>(alsk::arg::P<0>),
		Solution(alsk::arg::P<0>, alsk::arg::RNG),
		Solution(alsk::arg::R<0>)
	>,
	Solution(Solution, Solution)
>;

/* *** */

template<
	typename Problem, typename Solution,
	typename Init, typename LS, typename Select
>
using SkelGrasp = alsk::BuildSkeleton<SkelGraspStructure, SkelGraspLinks>::skeleton<
	tmp::Pack<Init, LS, Select>,
	tmp::Pack<Problem, Solution>
>;

}

#endif
