#ifndef ROSA_TABU_SEARCH_H
#define ROSA_TABU_SEARCH_H

#include <alsk/alsk.h>

namespace rosa {

template<
	typename InitLS,
	typename Copy,
	typename StopCriteria,
	typename FindNeighbour
>
using SkelTabuSearchStructure =
alsk::S<alsk::Serial,
	InitLS,
	Copy,
	alsk::S<alsk::While,
		StopCriteria,
		FindNeighbour
	>
>;

template<
	typename Problem, typename Solution
>
using SkelTabuSearchLinks =
alsk::L<alsk::Serial, void(Solution&, Problem const&),
	Solution(alsk::arg::P<1>, alsk::arg::RNG),
	void(alsk::arg::P<0>, alsk::arg::R<0>),
	alsk::L<alsk::While, void(alsk::arg::P<0>, alsk::arg::R<0>),
		bool(alsk::arg::P<0>),
		void(alsk::arg::P<0>, alsk::arg::P<1>)
	>
>;

/* *** */

struct Copy {
	template<typename T>
	void operator()(T& out, T const&in) const { out = in; }
};

template<
	typename Problem, typename Solution,
	typename InitLS, typename StopCriteria, typename FindNeighbour
>
using SkelTabuSearch = alsk::BuildSkeleton<SkelTabuSearchStructure, SkelTabuSearchLinks>::skeleton<
	tmp::Pack<InitLS, Copy, StopCriteria, FindNeighbour>,
	tmp::Pack<Problem, Solution>
>;

// template<
	// typename Problem, typename Solution, typename RNG,
	// typename InitLS, typename StopCriteria, typename FindNeighbour
// >
// using SkelTabuSearch =
// alsk::Serial<void(Solution&, Problem const&, RNG&),
	// alsk::Fun<InitLS, Solution(alsk::arg::P<1>, alsk::arg::P<2>)>,
	// alsk::Fun<Copy, void(alsk::arg::P<0>, alsk::arg::R<0>)>,
	// alsk::Fun<alsk::While<void(Solution&, Solution),
		// alsk::Fun<StopCriteria, bool(alsk::arg::P<0>)>,
		// alsk::Fun<FindNeighbour, void(alsk::arg::P<0>, alsk::arg::P<1>)>
	// >, void(alsk::arg::P<0>, alsk::arg::R<0>)>
// >;

}

#endif
