#ifndef TSP_POINT_H
#define TSP_POINT_H

#include <ostream>
#include <limits>

#include "types.h"

namespace tsp {

struct Point {
	Coord x;
	Coord y;

	Point(Coord x, Coord y): x{x}, y{y} {}

	friend bool operator==(Point const&p1, Point const&p2) {
		return
			std::fabs(p1.x-p2.x) < std::numeric_limits<Coord>::epsilon() &&
			std::fabs(p1.y-p2.y) < std::numeric_limits<Coord>::epsilon();
	}

	friend void swap(Point &p1, Point &p2) {
		std::swap(p1.x, p2.x);
		std::swap(p1.y, p2.y);
	}

	friend std::ostream &operator<<(std::ostream &os, Point const&p) {
		return os << p.x << "," << p.y;
	}
};

using Points          = std::vector<Point>;

}

#endif
