#include <algorithm>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <iterator>

#include "problem.h"
#include "solution.h"

namespace tsp {

void Solution::push_back(std::size_t i) {
	_value = valueWith(i);
	_indexes.push_back(i);
}

void Solution::swap(std::size_t i, std::size_t j) {
	std::size_t last = _indexes.size()-1;

	std::size_t pi = i == 0? last:i-1;
	std::size_t pj = j == 0? last:j-1;
	std::size_t ni = i == last? 0:i+1;
	std::size_t nj = j == last? 0:j+1;

	_value -= _problem->distances(_indexes[i], _indexes[pi]);
	_value -= _problem->distances(_indexes[i], _indexes[ni]);
	_value -= _problem->distances(_indexes[j], _indexes[pj]);
	_value -= _problem->distances(_indexes[j], _indexes[nj]);

	std::iter_swap(std::begin(_indexes)+i, std::begin(_indexes)+j);

	_value += _problem->distances(_indexes[i], _indexes[pi]);
	_value += _problem->distances(_indexes[i], _indexes[ni]);
	_value += _problem->distances(_indexes[j], _indexes[pj]);
	_value += _problem->distances(_indexes[j], _indexes[nj]);
}

void Solution::move2Opt(std::size_t e0, std::size_t b1) {
	assert(e0 < b1);

	_value = value2Opt(e0, b1);

	std::reverse(std::next(std::begin(_indexes), e0),
							 std::next(std::begin(_indexes), b1+1));
}

double Solution::valueWith(std::size_t i) const {
	std::size_t s = _indexes.size();
	Distance value = _value;
	if(s == 1)
		value = _problem->distances(_indexes[0], i) * 2;
	else if(s > 1) {
		value -= _problem->distances(_indexes[0], _indexes[s-1]);
		value += _problem->distances(_indexes[0], i);
		value += _problem->distances(_indexes[s-1], i);
	}
	return value;
}

double Solution::value2Opt(std::size_t e0, std::size_t b1) const {
	std::size_t const b0 = e0-1, e1 = (b1+1)%size();
	Distance value = _value;

	value -= _problem->distances(_indexes[b0], _indexes[e0]);
	value -= _problem->distances(_indexes[b1], _indexes[e1]);
	value += _problem->distances(_indexes[b0], _indexes[b1]);
	value += _problem->distances(_indexes[e0], _indexes[e1]);

	return value;
}

Solution Solution::offset(std::size_t o) const {
	Solution s;
	s._problem = _problem;
	s._value = _value;

	s._indexes.reserve(_indexes.size());
	for(std::size_t i = 0; i < _indexes.size(); ++i)
		s._indexes.push_back(_indexes[(i+o)%_indexes.size()]);

	return s;
}

Solution Solution::reverse() const {
	Solution s;
	s._problem = _problem;
	s._value = _value;

	s._indexes.reserve(_indexes.size());
	for(std::size_t i = 0; i < _indexes.size(); ++i)
		s._indexes.push_back(_indexes[_indexes.size()-i-1]);

	return s;
}

std::ostream &operator<<(std::ostream &os, Solution const&solution) {
	for(auto const&i: solution)
		os << solution._problem->points[i] << std::endl;
	return os;
}

void Solution::printIndices(std::ostream &os) const {
	std::copy(std::begin(_indexes), std::end(_indexes),
						std::ostream_iterator<decltype(_indexes)::value_type>(os, "\n"));
}

}
