#ifndef TSP_SOLUTION_H
#define TSP_SOLUTION_H

#include <iosfwd>
#include <limits>
#include <vector>

#include "types.h"

namespace tsp {

struct Problem;

class Solution {
public:
	using Indexes = std::vector<std::size_t>;

private:
	Problem const*_problem;
	Indexes _indexes;
	Distance _value;

public:
	Solution(): _problem{nullptr}, _value{std::numeric_limits<Distance>::infinity()} {}
	Solution(Problem const&problem): _problem{&problem}, _value{0} {}

	auto begin() const { return std::begin(_indexes); }
	auto rbegin() const { return std::rbegin(_indexes); }
	auto end() const { return std::end(_indexes); }
	auto rend() const { return std::rend(_indexes); }

	void push_back(std::size_t i);
	void swap(std::size_t i, std::size_t j);
	void move2Opt(std::size_t, std::size_t);

	Distance value() const { return _value; }
	Distance valueWith(std::size_t i) const;
	Distance value2Opt(std::size_t, std::size_t) const;

	bool empty() const { return _indexes.empty(); }
	std::size_t size() const { return _indexes.size(); }

	std::size_t operator[](std::size_t i) const { return _indexes[i]; }

	Solution offset(std::size_t) const;
	Solution reverse() const;

	friend bool operator==(Solution const&a, Solution const&b) { return a._indexes == b._indexes; }
	friend bool operator<(Solution const&a, Solution const&b) { return a._value < b._value; }

	friend std::ostream &operator<<(std::ostream &os, Solution const&solution);
	void printIndices(std::ostream &os) const;

};

}

#endif
